{% load i18n %}
<div class="failure sequence proctored-exam {{onboarding_status}}" data-exam-id="{{exam_id}}">
  <h3>
    {% blocktrans %}
      You must complete an onboarding exam before taking this proctored exam
    {% endblocktrans %}
  </h3>
  <p>
{% if onboarding_status == 'onboarding_missing' or onboarding_status ==  'onboarding_expired' %}
  <p>
    {% blocktrans %}
      Please complete an onboarding exam before attempting this exam.
    {% endblocktrans %}
  </p>
  {% if onboarding_link %}
    <a href="{{onboarding_link}}" class="btn-secondary">
      {% trans "Navigate to onboarding exam" %}
    </a>
  {% endif %}
{% elif onboarding_status == 'onboarding_pending' %}
  {% blocktrans %}
    Your onboarding exam is being reviewed. Before attempting this exam, please allow 2+ business days for your onboarding exam to be reviewed.
  {% endblocktrans %}
{% elif onboarding_status == 'onboarding_failed' %}
  <p>
    {% blocktrans %}
      Your onboarding exam failed to pass all requirements.
    {% endblocktrans %}
  </p>
  {% if onboarding_link %}
    <a href="{{onboarding_link}}" class="btn-secondary">
      {% trans "Navigate to onboarding exam" %}
    </a>
  {% endif %}
{% endif %}
  </p>
  <p>
    {% blocktrans %}
      Proctoring for your exam is provided via {{ provider_name }}.
      If you have questions about the status of your onboarding exam, contact
      <a href="mailto:{{ integration_specific_email }}">{{ integration_specific_email }}</a>.
    {% endblocktrans %}
  </p>
</div>
{% include 'proctored_exam/footer.html' %}
